package MoveNClick;

import TouchRecognizer.BlobDetector;
import TouchRecognizer.Point;
import java.awt.Toolkit;
import java.util.Collections;
import java.util.Vector;
import java.util.concurrent.Semaphore;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

/**
 *
 * @author dlm
 */
public class MoveNClick extends Thread
{
    // *** ATTRIBUTI ***
    private static MoveNClick istance = null;   // Istanza singelton
    private BlobDetector blobDetection;         // Oggetto che riconosce i blob a partire dall'immagine

    private Vector<Point> punti;                // Vettore che contiene i punti correnti
    private Vector<Event> eventi;               // Vettore che contiene gli eventi da passare al riconoscitore di zoom e rotazioni
    private static int nearThreshold;           // Soglia al di sopra della quale un punto si è spostato
    private static int sameThreshold;           // Soglia al di sotto della quale un punto è riconosciuto come lo stesso
    private Lock mutex;
    private Semaphore permessi;


    // *** METODI ***
    // Costruttore
    private MoveNClick()
    {
		try
		{
			this.punti = new Vector();
			this.eventi = new Vector();
			this.mutex = new ReentrantLock(true);
			this.permessi = new Semaphore(0);
			nearThreshold = sameThreshold = -1;
			// Avviamo il blob detector che a sua volta sarà un patten singleton che avvierà la webcam e setto gli attributi
			blobDetection = BlobDetector.getIstance();
			// Faccio partire il thread
			this.start();
		}
		catch(Exception ex)
		{
			System.out.println("Webcam non avviata: touch arrestato. " + ex);
		}
    }

    /**
     * Effettua il prodotto cartesiamo tra i punti rilevati a frame consecutivi
     * e li ordina in base alla distanza
     * @param Vettore di punti del frame attuale
     * @return Il vettore contenente coppie di punti ordinate per distanza crescente
     */
    private Vector<PairDistance> sortDistances(Vector<Point> puntiCorrenti)
    {
        Vector<PairDistance> distances = new Vector<PairDistance>(this.punti.size() * puntiCorrenti.size());

        // effettua il prodotto cartesiano dei punti gia rilevati con quelli correnti
        for(int i = 0; i < this.punti.size(); i++)
            for(int j = 0; j < puntiCorrenti.size(); j++)
                distances.add(new PairDistance(this.punti.elementAt(i), puntiCorrenti.elementAt(j), this.punti.elementAt(i).distance(puntiCorrenti.elementAt(j))));

        // ordina il vettore delle distanze
        Collections.sort((Vector)distances);

        return distances;
    }

    /*
     * Setta le soglie di lavoro
     */
    private void setThresholds()
    {
        // setta le soglie per discriminare press, move ecc.
        nearThreshold = (int) ((30 / (double) BlobDetector.getWidth()) * Toolkit.getDefaultToolkit().getScreenSize().width);
        sameThreshold = (int) ((5 / (double) BlobDetector.getWidth()) * Toolkit.getDefaultToolkit().getScreenSize().width);
        // propaga la soglia alla classe punto
        Point.setSameThreshold(sameThreshold);
    }
    
    @Override
    public void run()
    {
	//	try
	//	{
			while(true)
			{

				long millisecond = System.currentTimeMillis();
				// Processo l'immagine
				this.process();

				// alla prima esecuzione setta le soglie a seconda della risoluzione dell'immagine
				if(sameThreshold == -1)
					this.setThresholds();

				// rilascia il permesso sul vettore eventi
				if(this.permessi.availablePermits() == 0)
					this.permessi.release();

				// valuta il tempo impiegato per l'esecuzione
				millisecond = System.currentTimeMillis() - millisecond;

				// se impiega meno di 30 millisecondi
				// va in sleep per il tempo residuo
				// per ottenere una media di 30 fps
				try
				{
					if(millisecond < 30)
						Thread.sleep(30 - millisecond);
					else
						Thread.sleep(1);
				}
				catch(Exception e)
				{
					System.out.println(e);
				}
			}
	/*	}
		catch(Exception ex)
		{
			System.out.println("Touch arrestato: " + ex);
		}
    */}

    /**
     * Metodo che processa il vettore di punti ottenuto dal BlobDetector e crea i punti rilevati e il vettore degli eventi
     */
    private void process()
    {
        Vector<Point> puntiCorrenti = null;     // Vettore che contiene i punti trovati nel frame corrente
        Vector<PairDistance> distanze;          // Vettore che contiene le distanze tra coppie di punti
        Vector<Point> puntiEsaminati = new Vector<Point>();

        // garantisce la mutua esclusione dal vettore di eventi
        this.mutex.lock();

        // pulisce il vettore degli eventi
        this.eventi.clear();

        // Acquisisce i punti del frame corrente e crea il vettore temporaneo
        blobDetection.process();
        puntiCorrenti = this.blobDetection.getPoints();

        // ordina le coppie per distanze crescenti
        distanze = sortDistances(puntiCorrenti);

        // quindi scorre il vettore e abbina i punti più vicini
        boolean outOfBound = false;
        for(int i = 0; i < distanze.size() && !outOfBound; i++)
        {
            PairDistance tmp = distanze.get(i);

            // se non ha esaminato nessun punto della coppia significa che non sono stati abbinati
            // tra loro quindi non gli è stato associato un evento
            if(!(puntiEsaminati.contains(tmp.precedente) || puntiEsaminati.contains(tmp.attuale)))
            {
                // se lo spostamento è all'interno della soglia di vicinanza
                // allora i due punti sono correlati tra loro
                if(tmp.distanza < nearThreshold)
                {
                    // se il punto corrente è assimilabile al punto precedente
                    if(tmp.precedente.isTheSame(tmp.attuale))
                    {
                        // continua ad essere un press
                        // quindi aggiorna il punto e propaga l'evento
                        tmp.precedente.sposta(tmp.attuale);
                        this.eventi.addElement(new PressEvent(tmp.precedente.getX(), tmp.precedente.getY()));
                    }
                    // se invece il punto ha compiuto uno spostamento significativo
                    else
                    {
                        // è un move
                        // quindi propago l'evento
                        // ed aggiorno il punto
                        this.eventi.addElement(new MoveEvent(tmp.precedente.getX(), tmp.precedente.getY(),
                                                             tmp.attuale.getX(), tmp.attuale.getY()));
                        tmp.precedente.sposta(tmp.attuale);
                    }

                    // aggiunge i punti esaminati all'hashtable
                    // così che nei prossimi cicli non vengano più abbinati ad altri punti
                    puntiEsaminati.add(tmp.precedente);
                    puntiEsaminati.add(tmp.attuale);
                }
                // se la distanza è elevata, allora anche le coppie successive
                // non saranno correlate, quindi si può uscire dal ciclo
                else
                {
                    // esce dal ciclo
                    outOfBound = true;
                }
            }
        }

        // i punti disaccoppiati del frame precedente diventano click/release
        for(int i = 0; i < this.punti.size();)
        {
            // se il punto appartiene ai punti esaminati, significa
            // che è stato accoppiato, quindi non può essere un click
            if(puntiEsaminati.contains(this.punti.elementAt(i)))
            {
                // passa alla coppia successiva
                i++;
            }
            // altrimenti dovrà essere propagato un click/release
            else
            {
                // se lo spostamento assoluto è nullo significa che è un click
                if(this.punti.elementAt(i).getSpostamentoAssoluto() == 0)
                    // quindi viene propagato l'evento
                    this.eventi.addElement(new ClickEvent(this.punti.elementAt(i).getOriginX(), this.punti.elementAt(i).getOriginY()));
                // se invece se si è spostato è un release
                else
                    // quindi viene propagato 'evento
                    this.eventi.addElement(new ReleaseEvent(this.punti.elementAt(i).getX(), this.punti.elementAt(i).getY()));

                // viene rimosso il punto dal vettore
                this.punti.removeElement(this.punti.elementAt(i));
            }
        }

        // i punti non accoppiati rilevati nel nuovo frame diventano press
        for(int i = 0; i < puntiCorrenti.size(); i++)
        {
            // se il punto non è stato accoppiato è un press
            if(!puntiEsaminati.contains(puntiCorrenti.elementAt(i)))
            {
                // quindi viene propagato l'evento Press
                this.eventi.addElement(new PressEvent(puntiCorrenti.elementAt(i).getX(), puntiCorrenti.elementAt(i).getY()));
                // e aggiornato il vettore di punti
                this.punti.addElement(puntiCorrenti.elementAt(i));
            }
        }

        // viene rilasciato il lock sul vettore
        this.mutex.unlock();
    }

    /**
     * Metodo che ritorna il vettore degli eventi riconosciuti
     * @return Il vettore di eventi riconosciuti
     */
    public Vector<Event> getEvents()
    {
        Vector<Event> tmp = null;

        try
        {
            // riduce il numero di permessi sul vettore
            this.permessi.acquire();
            // e acquisisce il lock sul vettore per evitare inconsistenze
            this.mutex.lock();
            // per evitare che il chiamante possa modificare il vettore degli
            // eventi, ne viene passata una copia
            tmp = (Vector<Event>) this.eventi.clone();
            // rilascia il lock sul vettore
            this.mutex.unlock();
        }
        catch(Exception ex)
        {
            System.out.println("Esecuzione terminata: " + ex);
        }

        return tmp;
    }

    /**
     * Metodo che ritorna l'unica istanza dell'oggetto MoveNClick
     * @return L'istanza del MoveNClick
     */
    public static MoveNClick getIstance()
    {
        if(istance == null)
        {
            istance = new MoveNClick();
        }

        return istance;
    }

    private class PairDistance implements Comparable
    {
        Point precedente, attuale;
        Double distanza;

        public PairDistance(Point precedente, Point attuale, double distanza)
        {
            this.precedente = precedente;
            this.attuale = attuale;
            this.distanza = distanza;
        }

        public int compareTo(Object arg0)
        {
            int esito = -1;

            if(arg0 instanceof PairDistance)
            {
                PairDistance obj = (PairDistance) arg0;
                if(obj.distanza > this.distanza)
                    esito = -1;
                else if(obj.distanza < this.distanza)
                    esito = 1;
                else
                    esito = 0;
            }

            return esito;
        }
    }
}