package TouchRecognizer;

import java.awt.Toolkit;
import java.io.IOException;
import java.util.Vector;
import javax.media.CannotRealizeException;
import javax.media.NoPlayerException;

/**
 *
 * @author dlm
 */
public class BlobDetector
{
    // *** ATTRIBUTI ***
    private static final int MAX_POINT = 5;
    private static BlobDetector istance = null;
    private static int height = 0, width = 0;
    private CamManager cam;
    private byte[] rawData;
    private boolean[][] visitedPoint;
    private Vector<Point> points;
    private int areaThreshold;
    private boolean avgNvar; // true = average | false = variance
    private int avgLight;
    private int varLight;
    private int lightThreshold;
	private static int minX, minY, maxX, maxY;

    // *** METODI ***
    // Inizializza gli attributi.
    private BlobDetector() throws NoPlayerException, CannotRealizeException, IOException
    {
        points = new Vector<Point>();
        // al primo ciclo calcola la media
        avgNvar = true;
        cam = CamManager.getIstance();
        areaThreshold = 4;
        avgLight = 255;
        varLight = 255;
        lightThreshold = 0;
		minX = minY = 0;
		maxX = maxY = -1; // -1 sta ad indicare il massimo possibile
    }

    /*
     * Ottiene il punto x,y dall'array monodimensionale.
     * @param x Ascissa del punto interessato.
     * @param y Ordinata del punto interessato.
     * @return L'intensità luminosa del punto.
     */
    private int getPixel(int x, int y)
    {
        int pixel = 0;
        
        try
        {
            // controlla che il punto richiesto sia interno all'immagine
            if(x < maxX && y < maxY && x >= minX && y >= minY)
                pixel = (rawData[x + y * width]);
        }
        catch(Exception ex)
        {
            System.out.println(ex);
        }
        // per ovviare al problema dei byte con segno, se un punto è negativo
        // lo trasla per renderli tutti positivi
        return (pixel < 0) ? pixel + 256: pixel;
    }

    /*
     * Valuta se il punto x,y è premuto.
     * @param x Ascissa del punto interessato.
     * @param y Ordinata del punto interessato.
     * @return Vero se il punto è illuminato, falso altrimenti.
     */
    private boolean isPressed(int x, int y)
    {
        boolean esito = false;

        int pixel = getPixel(x,y);

        // controlla che il punto sia illuminato
        if(pixel > (this.avgLight + 2 * this.varLight + lightThreshold)
                && pixel > 20)
            esito = true;

        return esito;
    }

    /*
     * Cerca ricorsivamente una forma.
     * @param x Ascissa di partenza.
     * @param y Ordinata di partenza.
     * @param shape Forma a cui aggiungere eventualmente il punto.
     */
    private void findShape(int x, int y, Shape shape)
    {
        // se un punto è premuto e non ancora visitato
        // cerca nell'intorno per cercar di aggiungere altri punti
        if(isPressed(x,y) && !visitedPoint[x][y])
        {
            visitedPoint[x][y] = true;
            // aggrega questo punto alla forma
            shape.tryAttachPoint(new Point(x, y));

            // esplora ricorsivamente i pixel vicini
            if(x + 1 < width)
                findShape(x+1, y, shape);   // destra
            if(x > 0)
                findShape(x-1, y, shape);   // sinistra
            if(y > 0)
                findShape(x, y - 1, shape); // sopra
            if(y + 1 < height)
                findShape(x, y + 1, shape); // sotto;
        }
    }

    /*
     * Aggiunge i dati grezzi della webcam all'array monodimensionale.
     * Quindi setta le dimensioni dell'immagine e resetta la matrice
     * per il riconoscimento delle forme.
     */
    private void processCam()
    {
        this.rawData = null;

        while(rawData == null)
        {
            // ottiene i dati grezzi dalla webcam
            rawData = cam.getRawData();
            // ottiene la larghezza dell'immagine
            width = cam.getImageWidth();
            // ottiene l'altezza dell'immagine
            height = cam.getImageHeight();

			if(maxX == -1 || maxY == -1)
			{
				maxX = 320;
				maxY = 240;
			}
            // se i dati ricevuti dalla webcam non sono completi li invalidiamo
            if(rawData != null && rawData.length < width * height)
                rawData = null;
        }

        // inizializziamo la matrice dei punti visitati
        if(visitedPoint == null)
            visitedPoint = new boolean[width][height];
        for(int i = 0; i < width; i++)
            for (int j = 0; j < height; j++)
                visitedPoint[i][j] = false;   
    }

    /**
     * Ottiene la larghezza dell'immagine.
     * @return La larghezza dell'immagine della webcam.
     */
    public static int getWidth()
    {
        return width;
    }

    /**
     * Ottiene l'altezza dell'immagine.
     * @return L'altezza dell'immagine della webcam.
     */
    public static int getHeight()
    {
        return height;
    }

    /**
     * Analizza l'immagine e riconosce le forme.
     * I baricentri delle forme riconosciute vengono aggiunti a un vettore.
     */
    public void process()
    {
        long totalLight = 0;
        long totalLightVar = 0;

        // pulisce il vettore dei punti
        points.clear();

        // ottiene l'immagine dalla webcam
        processCam();

        // scorre i dati grezzi per riga
        for(int y = 0; y < maxY && points.size() < MAX_POINT; y++)
            for(int x = 0; x < maxX && points.size() < MAX_POINT; x++)
            {
                int pixel = getPixel(x,y);

                // a cicli alterni calcola media e varianza
                if(avgNvar)
                   totalLight += pixel;
                else
                    totalLightVar += Math.pow((pixel - this.avgLight), 2);

                // se il punto è illuminato e non ancora visitato
                // cerca di riconoscere la forma
                if(pixel > (this.avgLight + 2 * this.varLight + this.lightThreshold) &&
                   pixel > 20 &&
                   !visitedPoint[x][y])
                {
                    // riconosce un'intera forma
                    Shape shape = new Shape(width, height);
                    findShape(x,y, shape);

                    // aggiunge la forma alle forme riconosciute solo se l'area supera la soglia
                    if(shape.getArea() >= this.areaThreshold)
                    {
                        // adatta il baricentro alla risoluzione dello schermo
                        Point virtualBarycenter = shape.getBarycenter();
                        if(shape.isRound())
                        {
                            // inverte la posizione perchè la webcam ha l'immagine speculare
                            int realX = (int) (( (double)(width - 1 - virtualBarycenter.getX() - minX) / (maxX - minX - 1))
                                    * Toolkit.getDefaultToolkit().getScreenSize().width);
                            int realY = (int) (( (double)(height - 1 - virtualBarycenter.getY() - minY) / (maxY - minY - 1))
                                    * Toolkit.getDefaultToolkit().getScreenSize().height);

							points.add(new Point(realX, realY));
                        }
                    }
                }
            }

        if(avgNvar)
            // calcolo media
            this.avgLight = (int) (totalLight / (height * width));
        else
        {
            // calcolo varianza
            this.varLight = (int) Math.sqrt(totalLightVar / (height * width));
            // il punto deve essere del 20% oltre alla doppia deviazione standard
            this.lightThreshold = (int) (this.varLight * 0.2);
        }

        // il prossimo ciclo effettua l'inverso di quello che ha fatto in questo ciclo
        avgNvar = !avgNvar;
    }

    /**
     * Restituisce i baricentri delle forme riconosciute.
     * @return Un vettore di punti rappresentanti i baricentri delle forme.
     */
    public Vector<Point> getPoints()
    {
        return points;
    }

    /**
     * Ottiene l'unica istanza disponibile di BlobDetector. Se non ancora istanziata
     * richiama il costruttore privato così da inizializzare l'oggetto.
     * @return L'unica istanza disponibile di BlobDetector.
     */
    public synchronized static BlobDetector getIstance() throws NoPlayerException, CannotRealizeException, IOException
    {
        if(istance == null)
            istance = new BlobDetector();

        return istance;
    }

	public static void setMinX(int x)
	{
		minX = x;
                System.out.println("minX: " + minX);
	}
	
	public static void setMaxX(int x)
	{
		maxX = x;
                System.out.println("maxX: " + maxX);
	}

	public static void setMinY(int y)
	{
		minY = y;
                System.out.println("minY: " + minY);
	}

	public static void setMaxY(int y)
	{
		maxY = y;
                System.out.println("maxY: " + maxY);
	}
}
